#!/usr/bin/env python

'''
reproducer for (sporadic) rendering issues w/ Mako

(see: https://github.com/sqlalchemy/mako/issues/378)

To use, either:

- install python3.11 + pip (bug only seems to occur w/ python3.11)
- run pip(3) install -r requirements.txt
- run ./run-me.py


'''

import os
import sys
import urllib.parse
import logging

own_dir = os.path.dirname(__file__)
os.environ['CC_CONFIG_DIR'] = os.path.abspath(os.path.join(own_dir, 'cfg'))

from concourse.enumerator import (
    DefinitionDescriptorPreprocessor,
    GithubOrganisationDefinitionEnumerator,
    SimpleFileDefinitionEnumerator,
    TemplateRetriever,
)
from concourse.replicator import (
    FilesystemDeployer,
    PipelineReplicator,
    RenderOrigin,
    Renderer,
)

import ccc.concourse
import ci.log
import concourse.paths
import model.concourse
import model.github



class DummyDeployer:
    def deploy(self, definition_descriptor):
        pass # we do not care about storing the rendering output


def main(
    definition_file=os.path.join(own_dir, '.ci', 'pipeline_definitions'),
    cfg_name: str='external_active',
    repo_path: str = 'gardener/cc-utils',
    repo_branch: str = 'master',
    repo_host: str = 'github.com',
    template_path: str=concourse.paths.template_include_dir,
    template_include_dir: str=None,
):
    import ctx
    cfg_factory = ctx.cfg_factory()
    ci.log.configure_default_logging()

    cfg_set = cfg_factory.cfg_set(cfg_name=cfg_name)

    job_mapping = cfg_factory.job_mapping('external')
    secret_cfg = cfg_factory.secret('cicd-external')

    repo_url = 'https://github.com/gardener/cc-utils'

    def_enumerators = [
        SimpleFileDefinitionEnumerator(
            definition_file=definition_file,
            cfg_set=cfg_set,
            repo_path=repo_path,
            repo_branch=repo_branch,
            repo_host=repo_host,
            job_mapping=job_mapping,
            secret_cfg=secret_cfg,
        )
    ]

    preprocessor = DefinitionDescriptorPreprocessor()

    if not template_include_dir:
        template_include_dir = template_path

    template_retriever = TemplateRetriever(template_path=template_path)
    renderer = Renderer(
        template_retriever=template_retriever,
        template_include_dir=template_include_dir,
        cfg_set=cfg_set,
        render_origin=RenderOrigin.LOCAL,
    )

    deployer = DummyDeployer()

    replicator = PipelineReplicator(
        definition_enumerators=def_enumerators,
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer
    )

    import concurrent.futures
    p = concurrent.futures.ThreadPoolExecutor(max_workers=4)
    exceptions = []
    count = 0

    print('running template-renderings (will take some ten seconds)')

    def replicate():
        nonlocal exceptions
        nonlocal count
        try:
            replicator.replicate()
            count += 1
        except BaseException as e:
            print(f'exception at {count=}')
            exceptions.append(e)

    import logging
    jobs = []
    for i in range(256):
        jobs.append(p.submit(replicate))

    concurrent.futures.wait(jobs)

    print('done (there might be stacktraces above)')
    for e in exceptions:
        print(e)


if __name__ == '__main__':
    main()
